# -*- coding: utf-8 -*-
#
# Copyright (c) 2006-2007 by:
#     Blue Dynamics Alliance
#         * Klein & Partner KEG, Austria
#         * Squarewave Computing Robert Niederreiter, Austria
#
# GNU General Public License (GPL)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

__author__ = """Robert Niederreiter <rnix@squarewave.at>"""
__docformat__ = 'plaintext'

from zope.interface import implements
from zope.component import adapter
from zope.component import adapts

from bda.cache.interfaces import ICacheVary
from bda.cache.interfaces import ICacheManager
from bda.cache import FSCache
from bda.cache import CacheManager

from interfaces import ICacheVaryEvent

@adapter(ICacheVaryEvent)
def subscribeCacheVaryEvent(event):
    """Subscriber for the ICacheVaryEvent
    """
    fscache = FSCache(event.context['path'], createDirIfNotExist=True)
    cache = CacheManager(fscache)
    cache.setTimeout(event.context['timeout'])
    vary = ICacheVary(cache)
    vary.vary(*event.payload, **event.action)

class MailCacheModifiedEvent(object):
    """See interfaces.event.ICacheVaryEvent for details.
    """
    
    implements(ICacheVaryEvent)
    
    def __init__(self, context, action, payload):
        self.context = context
        self.action = { 'action': action, }
        self.payload = payload


class MailCacheVary(object):
    """See bda.cache.interfaces.ICacheVary for details.
    """
    
    implements(ICacheVary)
    adapts(ICacheManager)
        
    def __init__(self, context):
        self.cache = context
        self.actions = {
            'reload': self._reload,
            'flag': self._flag,
            'copy': self._copy,
            'move': self._move,
            'delete': self._delete,
            'reloadfolderlist': self._reloadFolderList,
        }
    
    def vary(self, *args, **kwargs):
        """This function tries to get the headers for the given folder and
        sets the given flag for message with id.
        """
        action = self.actions.get(kwargs['action'], None)
        if action:
            action(*args)
        else:
            raise
    
    def _reload(self, folder):
        """Delete headers of folder.
        """
        toreload = [
            '%s-headers' % folder,
            '%s-messagecount' % folder,
            '%s-unreadcount' % folder,
        ]
        for key in toreload:
            self.cache.rem(key)
    
    def _flag(self, folder, id, flag):
        """Set the flag for message header with id in folder.
        """
        if id is None:
            return
        
        key = '%s-headers' % folder
        headers = self.cache.get(key)
        
        if not headers:
            return
        
        for header in headers:
            if str(header.get('imapId', '')) == id:
                flags = header.get('flags', '')
                if flags.find(flag) == -1:
                    flags = '%s %s' % (flags, flag)
                    flags = flags.strip()
                    header['flags'] = flags
                    unreadkey = '%s-unreadcount' % folder
                    unread = self.cache.get(unreadkey)
                    unread -= 1
                    self.cache.set(unreadkey, unread)
                break
        
        self.cache.set(key, headers)
    
    def _reloadFolderList(self):
        """Simply deletes the folderlist entry from cache.
        """
        self.cache.rem('folderlist')
    
    def _copy(self, folder):
        """Remove cached headers for destinatoion folder from cache.
        """
        # TODO
        # implemented with reload atm
        pass
    
    def _move(self, folder, id, dest):
        """Delete header with id from folder headers and remove dest folder
        headers from cache.
        """
        # TODO
        # implemented with reload atm
        pass
    
    def _delete(self, folder, id):
        """Delete header with id from folder headers.
        """
        # TODO
        # implemented with reload atm
        pass
