# -*- coding: utf-8 -*-
#
# Copyright (c) 2006-2007 by:
#     Blue Dynamics Alliance
#         * Klein & Partner KEG, Austria
#         * Squarewave Computing Robert Niederreiter, Austria
#
# GNU General Public License (GPL)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

__author__ = """Robert Niederreiter <rnix@squarewave.at>"""
__docformat__ = 'plaintext'

"""
module WebMailControl.
"""

import types
import time
import logging
logger = logging.getLogger('BlueWebMail.control')

from zope.app import zapi
from zope.interface import implements
from zope.component import getAdapter
from zope.component import getMultiAdapter
from zope.event import notify

from bda.email.imap.interfaces import IIMAPProperties
from bda.email.imap.interfaces import IIMAPServerSettings
from bda.email.imap import Gateway
from bda.email.imap import Mailbox

from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.browser.interfaces import IPlone

from interfaces import IFsCacheProperties
from manufacture import Manufacture
from cachevary import MailCacheModifiedEvent
from util import setWebmailMessage

def getMailboxForAuthenticatedUser(context):
    """Get the Mailbox for the current authenticated user
    
    Try to find an object of type WebMailBox in the portal catalog created
    by the current authenticated user.
    
    If more then one is found, the first one is returned as 2-tuple containing
    (mailboxobject, 'urltomailboxobject').
    
    If no one is found return None.
    """
    pm = getToolByName(context, 'portal_membership')
    currentUser = pm.getAuthenticatedMember().getId()
    catalog = getToolByName(context, 'portal_catalog')
    brains = catalog(portal_type='WebMailBox',
               Creator=currentUser)
    if brains:
        return (brains[0].getObject(), brains[0].getURL())
    return None

class Control(object):
    """Controller Object for WebMail Product.
    
    This object is responsible for all manipulation tasks of the mailbox
    since in my opinion the so called 'view classes' imply the term view in
    the sense of 'model - view - control' and must not need know about
    the business logic of a software.
    """
    
    def __init__(self, context):
        """Init the controller Object.
        """
        self.context = context
        self.request = None
        
        logger.info('Create Control object')
        
        # modify this fpr direct use of the cache properties
        cacheprops = IFsCacheProperties(context)
        cachedefs = None
        if cacheprops.cache_enabled:
            cachedefs = dict()
            cachedefs['path'] = cacheprops.cachedir
            cachedefs['timeout'] = cacheprops.timeout
        self._cachedefs = cachedefs
        
        
        imapprops = IIMAPProperties(context)
        self._serversettings = IIMAPServerSettings(context)
        gate = Gateway(imapprops)
        self._mb = Mailbox(gate, self._serversettings, cachedefs)
        self._mb.connect()
        self._mb.login() # TODO: readonly handling
        self._mb.checkDefaultFolders()
        
    def initialize(self, context, request):
        """Initialize the current state of the controller object.
        """
        # reset context
        self.context = context
        self.request = request
        
        if not self._mb.check():
            self._mb.connect()
            self._mb.login() # TODO: readonly handling
        self.request = request
        self._execute()
    
    def skip(self):
        return self.request.get('skip_remaining', False)
    
    @property
    def mailbox(self):
        """Return the mailbox object. needed in the view classes for displaying
        the mailbox contents.
        """
        return self._mb
    
    def _execute(self):
        """Execute the business logic.
        
        order:
            * check if creation mode has to be changed
            * check the actions new, reply, reply all and forward
            * check actions for message creation
            * check message management actions
            * check folder management actions
        
        If one of them functions promise that all neccesary tasks are done for
        this request, this function returns premature.
        """
        functions = [self._checkReload,
                     self._checkFlag,
                     self._switchCreationMode,
                     self._redirectIfCreation,
                     self._checkCreateMessageActions,
                     self._checkMessageManagementActions,
                     self._checkFolderManagementActions,]
        
        for func in functions:
            func()
            if self.skip():
                return
    
    def _checkReload(self):
        """Check if reload was called.
        """
        if self.request.get('refresh', None):
            folder = self.request.get('folder', 'INBOX')
            event = MailCacheModifiedEvent(self._cachedefs, 'reload', [folder])
            notify(event)
    
    def _checkFlag(self):
        """Check if cached header has to be marked as \Seen.
        """
        plone = getMultiAdapter((self.context, self.request), name=u'plone')
        cur_url = plone.getCurrentUrl()
        if cur_url.find('message_view') != -1:
            folder = self.request.get('folder', 'INBOX')
            id = self.request.get('imapId', None)
            flag = '\Seen'
            args = [folder, id, flag]
            event = MailCacheModifiedEvent(self._cachedefs, 'flag', args)
            notify(event)
    
    def _switchCreationMode(self):
        """Check if the creation mode has to be changed.
        """
        request = self.request
        context = self.context
        if request.get('switchToText', None):
            self.context.setCreateFormat('TEXT')
        if request.get('switchToHtml', None):
            self.context.setCreateFormat('HTML')
    
    def _checkMessageManagementActions(self):
        """Do not change the order of this checks.
        """
        r = self.request
        defaults = self._mb.default_folders
        folder = self.request.get('folder', defaults['inbox'])
        f = self._mb.getFolder(folder)
        
        messageView = False
        if r.get('PATH_TRANSLATED').find('message_view') > -1:
            messageView = True  
        p = None
        n = None
        
        # previous and next Message actions
        if messageView:
            id = r.get('imapId', '').strip()
            sort = r.get('sort', '')
            message = f.getMessages([id])
            p, n = message[0].getPrevAndNextId(sort)
            r['showPrev'] = True
            r['showNext'] = True
            prev = r.get('previous')
            next = r.get('next')
            # check if previous or next message 
            if prev is not None or next is not None:
                if prev is not None:
                    id = str(p)
                elif next is not None:
                    id = str(n)
                r['imapId'] = id
        
        # delete message(s)        
        if r.get('delete', '').strip() != '':
            imapIds = r.get('imapId', '')
            if imapIds != '':
                if type(imapIds) is types.StringType:
                    imapIds = [imapIds]   
                messages = f.getMessages(imapIds)
                setWebmailMessage(self.request, f.deleteMessages(messages))
                eventfolder = [r.get('folder', defaults['inbox'])]
                event = MailCacheModifiedEvent(self._cachedefs,
                                               'reload',
                                               eventfolder)
                notify(event)
                if messageView:
                    if n is None and p is not None:
                        id = str(p)
                    elif p is None and n is not None:
                        id = str(n)
                    elif n is not None and p is not None:
                        id = str(n)
                    else:
                        id = ''
                    r['imapId'] = id
        
        # move message(s)
        if r.get('move', '').strip() != '':
            imapIds = r.get('imapId', '')
            folderDest = r.get('folderDest', '').strip()
            currentFolder = r.get('folder', defaults['inbox'])
            moveToSelf = currentFolder == folderDest
            if moveToSelf:
                setWebmailMessage(self.request, 'move_to_self_not_allowed')
            if imapIds != '' and folderDest != '' and not moveToSelf:
                if type(imapIds) is types.StringType:
                    imapIds = [imapIds]
                messages = f.getMessages(imapIds)
                setWebmailMessage(self.request,
                                  f.moveMessages(folderDest, messages))
                event = MailCacheModifiedEvent(self._cachedefs,
                                               'reload',
                                               [currentFolder])
                notify(event)
                event = MailCacheModifiedEvent(self._cachedefs,
                                               'reload',
                                               [folderDest])
                notify(event)
                if messageView:
                    if n is None and p is not None:
                        id = str(p)
                    elif p is None and n is not None:
                        id = str(n)
                    elif n is not None and p is not None:
                        id = str(n)
                    else:
                        id = ''
                    r['imapId'] = id  
        
        # copy message(s)
        if r.get('copy', '').strip() != '':
            imapIds = r.get('imapId', '')
            folderDest = r.get('folderDest', '').strip()
            if imapIds != '' and folderDest != '':
                if type(imapIds) is types.StringType:
                    imapIds = [imapIds]
                messages = f.getMessages(imapIds)
                setWebmailMessage(self.request,
                                  f.copyMessages(folderDest, messages))
                event = MailCacheModifiedEvent(self._cachedefs,
                                               'reload',
                                               [folderDest])
                notify(event)
        
        #define prev and next button display condition
        if messageView:
            if id is not None:
                message = f.getMessages([id])
                p, n = message[0].getPrevAndNextId(sort)
                if p is None:
                    r['showPrev'] = False
                if n is None:
                    r['showNext'] = False
    
    def _checkFolderManagementActions(self):
        """Check if a folder has to be created or deleted.
        """
        r = self.request
        base = self.context.absolute_url()
        
        currentfolder = r.get('folder')
        createfolder = r.get('createfolder', None)
        deletefolder = r.get('deletefolder', None)
        
        # nothing to do
        if createfolder == None and deletefolder is None:
            return
        
        if createfolder is not None:
            currentfolder = currentfolder.strip()
            self._createFolder(currentfolder, createfolder)
            
        if deletefolder is not None:
            deletefolder = deletefolder.strip()
            self._deleteFolder(deletefolder)
        
        url = '%s?folder=%s' % (base, currentfolder)
        r.response.redirect(url)
        r['skip_remaining'] = True
    
    def _checkCreateMessageActions(self):
        """Check if a message was created and has to be saved or sent.
        """
        # ugly workaround since there are problems with redirecting <-- ??
        r = self.request
        if r.get('send', '').strip() != '':
            manufacture = Manufacture(self.context, r, Manufacture.ACTION_SEND)
            manufacture.perform(self._mb)
        
        if r.get('save', '').strip() != '':
            manufacture = Manufacture(self.context, r, Manufacture.ACTION_SAVE)
            manufacture.perform(self._mb)
        
    def _redirectIfCreation(self):
        """Redirect to create page if there is corresponding action.
        """
        r = self.request
        
        f = r.get('folder', '')
        id = r.get('imapId', '')
        
        url = '%s/create_message' % self.context.absolute_url()
        
        if r.get('new', None) is not None:
            r.response.redirect('%s?createAction=New&folder=%s' % (url, f))
            r['skip_remaining'] = True
        if r.get('reply', None) is not None:
            query = 'createAction=Reply&folder=%s&imapId=%s' % (f, id)
            r.response.redirect('%s?%s' % (url, query))
            r['skip_remaining'] = True
        if r.get('reply_all', None) is not None:
            query = 'createAction=Reply All&folder=%s&imapId=%s' % (f, id)
            r.response.redirect('%s?%s' % (url, query))
            r['skip_remaining'] = True
        if r.get('forward', None) is not None:
            query = 'createAction=Forward&folder=%s&imapId=%s' % (f, id)
            r.response.redirect('%s?%s' % (url, query))
            r['skip_remaining'] = True
    
    def _deleteFolder(self, foldername):
        """Delete folder.
        
        This Function should be abstracted and placed inside the mailbox object
        """
        defaults = self._mb.default_folders
        defaults = [defaults[key] for key in defaults.keys()]
        
        if foldername in defaults:
            setWebmailMessage(self.request, 'cannot_delete_default_folder')
        
        if foldername == '':
            setWebmailMessage(self.request, 'no_folder_name')
        else:
            foldernames = self._mb.listFolders()
            foldernames = [fname[1] for fname in foldernames]
            todelete = []
            for fname in foldernames:
                if fname.startswith(foldername):
                    todelete.append(fname)
            for fname in todelete:
                folder = self._mb.getFolder(fname)
                ret = folder.delete()
                if ret[0] == 'NO':
                    setWebmailMessage(self.request, ret[1][0])
                else:
                    fname = fname[fname.rfind('.') + 1:]
                    setWebmailMessage(self.request,
                                      'folder_deleted',
                                      mapping={'folder': fname})
                
                event = MailCacheModifiedEvent(self._cachedefs,
                                               'reloadfolderlist',
                                               [])
                notify(event)
    
    def _createFolder(self, currentfolder, createfolder):
        """Create folder in current folder.
        
        This function should be abstracted and placed inside the mailbox object
        """
        delim = self._mb.delimiter
        prefix = self._mb.folder_prefix
        inbox = self._mb.default_folders['inbox']
        default_sub = self._mb.default_sub_of_inbox
        encoding = self._mb.folder_creation_encoding
        
        if createfolder == '':
            setWebmailMessage(self.request, 'no_folder_name')
            return
        
        if createfolder.find(delim) != -1:
            setWebmailMessage(self.request, 'invalid_folder_name')
            return
        
        try:
            createfolder.decode(encoding)
        except UnicodeDecodeError:
            setWebmailMessage(self.request, 'invalid_folder_name')
            return
        
        createfolder = createfolder.replace(' ', '_')
        if default_sub:
            # TODO: handle default sub of inbox
            pass
        else:
            if currentfolder == inbox:
                foldername = '%s%s' % (prefix, createfolder)
            else:
                foldername = '%s%s%s' % (currentfolder,
                                         delim,
                                         createfolder)
            
            foldername = foldername.decode(encoding)
            folder = self._mb.getFolder(foldername)
            res = folder.create()
        
            if res == 'NO':
                setWebmailMessage(self.request, 'cannot_create_folder')
            else:
                setWebmailMessage(self.request,
                                  'folder_created',
                                  mapping={'folder': createfolder})
                
                event = MailCacheModifiedEvent(self._cachedefs,
                                               'reloadfolderlist',
                                               [])
                notify(event)
    