# -*- coding: utf-8 -*-
#
# Copyright (c) 2006-2007 by:
#     Blue Dynamics Alliance
#         * Klein & Partner KEG, Austria
#         * Squarewave Computing Robert Niederreiter, Austria
#
# GNU General Public License (GPL)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

__author__ = """Robert Niederreiter <rnix@squarewave.at>"""
__docformat__ = 'plaintext'

"""Module keeper.

This module contains a global named 'tool' which is instantiated at init time
of the BlueWebMail Product with a Keeper object.

This Keeper object is a singleton and responsible for controller objects
bound to session id's.
"""

import time
import threading
import logging

from control import Control

tool = None
logger = logging.getLogger('BlueWebMail.keeper')

class Keeper(object):
    """Class Keeper is responsible for keeping Controller Objects which
    themselves keep the connection to the IMAP Server.
    
    This is done by pseudo binding to the Session and holding the corresponding
    Controller object in this Object.
    """
    
    checkintervall = 10
    lastcheck = 0
    
    def __init__(self):
        """Construct the object holding the Controllers.
        
        The container for the controllers is an OOBTree and its contents have
        the following structure:
        {
            'sessionid': {
                'sessionid'
                'lastaccessed': timestamp
                'controller': control.Control
            }
        }
        """
        logger.info('initializing Keeper object')
        self.controllers = dict()
        self.lock = threading.RLock()
    
    def getController(self, context, request, timeout):
        """Return either a new or an already existing controller object.
        
        Take a request out of which the session data is extracted and a
        timeout in minutes
        """
        benchstart = time.time()
        self.lock.acquire()
        try:
            timeout = timeout * 60
            currenttime = time.time()
            sessionid = request.SESSION.getId()
            
            controller = self.controllers.get(sessionid, None)
            if controller:
                controller = controller['controller']
                self.controllers[sessionid]['lastaccessed'] = currenttime
            else:
                controller = Control(context)
                self.controllers[sessionid] = dict()
                self.controllers[sessionid]['sessionid'] = sessionid
                self.controllers[sessionid]['lastaccessed'] = currenttime
                self.controllers[sessionid]['controller'] = controller
            
            if currenttime - self.checkintervall > self.lastcheck:
                self._kickTimedOut(currenttime, timeout) 
                self.lastcheck = currenttime
            
            initialized = request.get('webmail_initialized', False)
            if not initialized:
                controller.initialize(context, request)
                request['webmail_initialized'] = True
            
            return controller
        finally:
            self.lock.release()
            benchend = time.time()
        
    def _kickTimedOut(self, currenttime, timeout):
        """Checks for controllers which session has timed out.
        """
        try:
            to_delete = []
            for id in self.controllers:
                if currenttime - timeout > self.controllers[id]['lastaccessed']:
                    to_delete.append(id)
            for id in to_delete:
                ctrl = self.controllers.pop(id)
                ctrl['controller'].mailbox.logout()
        except RuntimeError, rte:
            logger.error(str(rte))
