# -*- coding: utf-8 -*-
#
# Copyright (c) 2006-2007 by:
#     Blue Dynamics Alliance
#         * Klein & Partner KEG, Austria
#         * Squarewave Computing Robert Niederreiter, Austria
#
# GNU General Public License (GPL)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

__author__ = """Robert Niederreiter <rnix@squarewave.at>"""
__docformat__ = 'plaintext'

import smtplib
import types
from sets import Set
from email import Encoders
from email.Header import Header
from email.MIMEMultipart import MIMEMultipart
from email.MIMEAudio import MIMEAudio
from email.MIMEBase import MIMEBase
from email.MIMEImage import MIMEImage
from email.MIMEText import MIMEText
from email.Utils import getaddresses
from email.Utils import formatdate

from zope.interface import implements
from zope.component import adapts
from zope.component import getAdapter
from zope.event import notify

from Products.CMFCore.utils import getToolByName
from Products.validation import validation
from Products.MimetypesRegistry.mime_types.magic import guessMime

from interfaces import IMessage
from interfaces import IMailSendout
from interfaces import IReceiverResolver
from interfaces import IReceiverProvider
from interfaces import IReceiverRecord
from interfaces import IDispatchRuleSet
from interfaces import ISMTPProperties
from interfaces import IWebmailProperties
from util import setWebmailMessage
from cachevary import MailCacheModifiedEvent


class SendingFailed(Exception): pass


class Manufacture(object):
    """Class Manufacture is the controller class for creating, storing and
    sending email messages.
    """
    
    ACTION_SEND = 1
    ACTION_SAVE = 2

    def __init__(self, context, request, action):
        self.context = context
        self.request = request
        self.action = action
        
    def perform(self, mailbox):
        """Perform this request.
        
        Take a bda.email.imap.IMAPMailBox as argument.
        
        Create a message Factory, get the message object out of it and store it
        to self. Then check the action to perform (save or send the message).
        Finally either store the message to the drafts or send folder,
        depending on the current action. This is done by the given mailbox
        object.
        """
        props = IWebmailProperties(self.context)
        
        fromaddr = '%s <%s>' % (props.real_name, props.email_address)
        mf = MessageFactory(self.request)
        mf.initialize(fromaddr, mailbox, props)
        self.msg = mf.getMessage()
        defaults = mailbox.default_folders
        
        if self.action == self.ACTION_SEND:
            folder = mailbox.getFolder(defaults['sent'])
            folder.insertMessages([self.msg])
            if mailbox.cachedefs:
                event = MailCacheModifiedEvent(
                    mailbox.cachedefs, 'reload',
                    [mailbox.default_folders['sent']])
                notify(event)
            self._sendMessage()
        
        if self.action == self.ACTION_SAVE:
            folder = mailbox.getFolder(defaults['drafts'])
            if mailbox.cachedefs:
                event = MailCacheModifiedEvent(
                    mailbox.cachedefs, 'reload',
                    [mailbox.default_folders['drafts']])
                notify(event)
            self._saveMessage(folder)
    
    def _saveMessage(self, folder):
        """Save the message.
        """
        createAction = self.request.get('createAction', '')
        if createAction == 'Draft':
            imapId = self.request.get('imapId', '')
            if not imapId:
                raise
            message = folder.getMessages([imapId])
            if not message:
                raise
            message[0].deleteMessage()
            folder.insertMessages([self.msg])
        else:
            folder.insertMessages([self.msg])
        
        self._redirect('message_saved_to_drafts_folder')
        
    def _sendMessage(self):
        """Send the message.
        """
        receiverResolver = IReceiverResolver(self.msg)
        
        error = False
        invalid = []
        
        return_msg = 'message_sent'
        mapping = None
        
        try:
            message = receiverResolver.resolve(data=self.context)
            invalid = receiverResolver.getInvalid()
            sendout = IMailSendout(message)
            sendout.initialize(data=self.context)
            sendout.send()
        except SendingFailed, e:
            error = True
            return_msg = e.args[0]
            mapping = e.args[1]
        except Exception, e:
            error = True
            return_msg = 'sending_failed'
            mapping = { 'reason': str(e) }
        
        if invalid and not error:
            invalid = ', '.join(invalid)
            return_msg = 'invalid_receips'
            mapping = { 'invalid': invalid }
            
        self._redirect(return_msg, mapping)
    
    def _redirect(self, message, mapping=None):
        """Redirect after saving is done.
        
        TODO: some error handling if saving fails.
        """
        setWebmailMessage(self.request, message, mapping)
        f = self.request.get('folder', '')
        base = self.context.absolute_url()
        url = '%s/?folder=%s' % (base, f)
        self.request.response.redirect(url)
        self.request['skip_remaining'] = True


class Sendout(object):
    """See Products.BlueWebMail.interfaces.IMailSendout for details.
    """
    
    implements(IMailSendout)
    adapts(IMessage)
    
    def __init__(self, context):
        self.context = context
        
    def initialize(self, data=None):
        """Initialite this Sendout object.
        
        In our case we take an IWebMail implementing object as data argument
        to look up the SMTP server preferences by this object.
        """
        message = self.context
        self.receivers = Set()
        
        to = message.get_all('to', [])
        cc = message.get_all('cc', [])
        bcc = message.get_all('bcc', [])
        all = getaddresses(to + cc + bcc)
        
        for address in all:
            if address[1] == '':
                continue
            self.receivers.add(address[1])
        
        message.__delitem__('Bcc')
        
        self.smtpprefs = ISMTPProperties(data)
    
    def send(self):
        """Send the message.
        
        This Sendout object uses the standard python smtplib.
        """
        smtpprefs = self.smtpprefs
        server = smtplib.SMTP(smtpprefs.server, smtpprefs.port)
        if smtpprefs.auth_required:
            server.login(smtpprefs.user, smtpprefs.passwd)
        server.set_debuglevel(0)
        message = self.context
        
        # in case of groupmailing there might be no 'to' header set
        settoheader = False
        if not message.get('To'):
            settoheader = True
        
        for receiver in self.receivers:
            if settoheader:
                message.__delitem__('To')
                message.add_header('To', receiver)
            server.sendmail(message['From_'], receiver, message.as_string())
        server.quit()


class ReceiverResolver(object):
    """See Products.BlueWebMail.interfaces.IReceiverResolver for details.
    """
    
    implements(IReceiverResolver)
    adapts(IMessage)
    
    def __init__(self, context):
        self.context = context
        self.data = None
        self.invalid = []
    
    def resolve(self, data=None):
        self.data = data
        
        to, to_invalid = self._getReceivers('To')
        cc, cc_invalid = self._getReceivers('Cc')
        bcc, bcc_invalid = self._getReceivers('Bcc')
        
        if not to and not cc and not bcc:
            raise SendingFailed('no_receips_given', None)
        
        self._defineInvalid(to_invalid + cc_invalid + bcc_invalid)
        self._alterHeaders(to, cc, bcc)
                
        dispatchruleset = IDispatchRuleSet(self.context)
        
        if not dispatchruleset.appliesToRuleSet(data=data):
            limit = str(dispatchruleset.getCurrentLimit())
            count = str(dispatchruleset.getCurrentCount())
            mapping = { 'count': count, 'limit': limit }
            raise SendingFailed('invalid_receips_count', mapping)
        
        return self.context
    
    def getInvalid(self):
        return self.invalid
    
    def _defineInvalid(self, invalid):
        set = Set()
        for address in invalid:
            set.add(address)
        invalid = [address for address in invalid if address]
        self.invalid = list(set)
    
    def _getReceivers(self, header):
        resolved = []
        receivers = self._stripHeader(header)
        
        for rec in receivers:
            if rec[1].find(':') > -1:
                name = '.%s' % rec[1][rec[1].find('<') + 1:rec[1].find(':')]
                recordname = rec[1][rec[1].find(':') + 1:rec[1].find('>')]
                
                record = ReceiverRecord(recordname)
                receiverProvider = getAdapter(record, IReceiverProvider, name)
                receiverProvider.process(data=self.data)
                
                recordreceivers = record.getReceivers()
                recordreceivers = getaddresses([','.join(recordreceivers)])
                
                for recordreceiver in recordreceivers:
                    resolved.append(recordreceiver[1])
            else:
                resolved.append(rec[1])
            
        return self._validateReceivers(resolved)
    
    def _stripHeader(self, header):
        header = self.context.get(header, '')
        header = header.split(',')
        addresses = []
        groups =  []
        
        for address in header:
            if address.find(':') != -1:
                groups.append(('', address))
            else:
                addresses.append(address)
                
        addresses = [','.join(addresses)]
        addresses = getaddresses(addresses)
        
        ret = addresses + groups
        return ret
    
    def _validateReceivers(self, receivers):
        validated = []
        failed = []
        
        for receiver in receivers:
            if validation.validate('isEmail', receiver) == 1:
                validated.append(receiver)
            else:
                failed.append(receiver)
        
        return validated, failed
        
    def _alterHeaders(self, tos, ccs, bccs):
        headers = (('To', tos), ('Cc', ccs), ('Bcc', bccs))
        
        for header in headers:
            headerstring = ''
            
            for address in header[1]:
                headerstring = '%s%s, ' % (headerstring, address)
            
            headerstring = headerstring[:headerstring.rfind(',')]
            self.context.__delitem__(header[0])
            self.context.add_header(header[0], headerstring)
            

class ReceiverRecord(object):
    """See Products.BlueWebMail.interfaces.IReceiverRecord for details.
    """
    
    implements(IReceiverRecord)
    
    def __init__(self, recordname):
        self.recordname = recordname
        self.receivers = []
    
    def getRecordName(self):
        return self.recordname
    
    def setReceivers(self, receivers):
        self.receivers = receivers
    
    def getReceivers(self):
        return self.receivers


class DispatchRuleSet(object):
    """See interfaces.IDispatchRuleSet for details.
    """
        
    implements(IDispatchRuleSet)
    adapts(IMessage)
    
    def __init__(self, context):
        self.context = context
        self.currentlimit = -1
    
    def appliesToRuleSet(self, data=None):
        limit = self._getLimitForCurrentUser(data)
        count = self._countReceivers()
        self.currentlimit = limit
        self.currentcount = count
        
        if limit == -1 or count <= limit:
            return True
        return False
    
    def getCuttentLimit(self):
        return self.currentlimit
    
    def getCuttentCount(self):
        return self.currentcount
    
    def _getLimitForCurrentUser(self, data):
        """Return the lowest limit for the current authenticated user.
        """
        wmt = getToolByName(data, 'portal_webmailtool')
        limitstrings = wmt.getGroupMessageLimit()
        limits = []
        
        for limit in limitstrings:
            current = limit.split(':')
            current = [e.strip() for e in current]
            limits.append((current[0], int(current[1])))
        
        groups = self._getGroupsForMember(data)
        
        # mailsending limits, order is important. wildcards like 'foo*bar' are
        # not supported
        for limit in limits:
            
            # wildcard handling
            if limit[0].find('*') != -1:
                
                # all groups are affected
                if limit[0] == '*':
                    return limit[1]
                
                # wildcard matches like '*foo'
                elif limit[0].startswith('*'):
                    for group in groups:
                        if group.endswith(limit[0][1:]):
                            return limit[1]
                
                # wildcard matches like 'foo*'
                elif limit[0].endswith('*'):
                    for group in groups:
                        if group.startswith(limit[0][:-1]):
                            return limit[1]
                
                # wildacard matches like '*foo*'
                else:
                    for group in groups:
                        if group.find(limit[0][1:-1]) != -1:
                            return limit[1]
            
            # exact match
            else:
                for group in groups:
                    if group == limit[0]:
                        return limit[1]
        
        # nothing was effected or no limitiations made, unlimited sending is
        # allowed
        return -1

    def _getGroupsForMember(self, data):
        """Return a list groupids the current authenticated user is member of.
        """
        pm = getToolByName(data, 'portal_membership')
        user = pm.getAuthenticatedMember().getId()
        grouptool = getToolByName(data, 'portal_groups')
        groups = grouptool.getGroupsByUserId(user)
        ret = []
        
        for group in groups:
            ret.append(group.getId())
        return ret
    
    def _countReceivers(self):
        message = self.context
        to = message.get_all('to', [])
        cc = message.get_all('cc', [])
        bcc = message.get_all('bcc', [])
        all = getaddresses(to + cc + bcc)
        
        unique = Set()
        for address in all:
            if address[1] == '':
                continue
            unique.add(address)
        
        return len(unique)


class MessageFactory(object):
    """Factory for E-Mail Message Objects.
    """
    
    def __init__(self, request):
        self.request = request
        self.message = None
    
    def initialize(self, fromaddress, mailbox, props):
        """Read the message data from request and create an email.
        """
        req = self.request
        message = None
        
        charset = 'utf-8'
        attachments = req.get('append-attachment', [])
        bodyText = req.get('messageText', '')
        forward = req.get('forwardmessage', None)
        fromDrafts = req.get('createAction', '') == 'Draft'
        
        # case sending from drafts. we want to take care of the already
        # existing attachments
        draftattachments = []
        if fromDrafts:
            folder = mailbox.getFolder(req.get('folder'))
            draftmessage = folder.getMessages([req.get('imapId', '')])
            if not draftmessage:
                raise
            parts = draftmessage[0].getParts()
            for part in parts:
                if part.attachment:
                    draftattachments.append(part.part)
        
        subtypes = {
            'TEXT': 'plain',
            'HTML': 'html',
        }
        subtype = subtypes[props.create_format]
        
        if not attachments and not forward and not draftattachments:
            message = MIMEText(bodyText, _subtype=subtype)
            message.set_charset(charset)
        else:
            mh = MimeHelper()
            message = MIMEMultipart()
            message.set_charset(charset)
            mailbody = mh.wrapMimeMessage('text/%s' % subtype, bodyText)
            mailbody.set_charset(charset)
            message.attach(mailbody)
            
            if forward:
                folder = mailbox.getFolder(req.get('folder'))
                messages = folder.getMessages([req.get('imapId', '')])
                if not messages:
                    raise
                attachment = messages[0].getMessage()
                message.attach(attachment)
            
            if fromDrafts and draftattachments:
                for attachment in draftattachments:
                    message.attach(attachment)
            
            if not isinstance(attachments, list):
                attachments = [attachments]
            for a in attachments:
                data = a.read()
                if not data:
                    continue
                
                # irgendwas geht da noch nicht
                # attachment.set_param('name', a.filename)
                
                attachment = mh.wrapMimeMessage(None, data)
                attachment.add_header('Content-Disposition',
                                      'attachment', 
                                      filename=a.filename)
                message.attach(attachment)
        
        subject = req.get('subject', '')
        subject = subject.decode('utf-8')
        subject = Header(subject, 'utf-8')
        
        message.add_header('Date',  formatdate(localtime=True))
        message.add_header('From_', fromaddress)
        message.add_header('From', fromaddress)
        message.add_header('To', req.get('to', ''))
        message.add_header('Cc', req.get('cc', ''))
        message.add_header('Bcc', req.get('bcc', ''))
        message['Subject'] = subject
        
        self.message = message
    
    def getMessage(self):
        """return the message.
        """
        return self.message


class MimeHelper(object):
    """Helper Class for MIME Messages.
    """
    
    def wrapMimeMessage(self, contenttype, body):
        """Take content type and payload for this message part, wrap it
        into a MIMEAny object and return it.
        """
        
        if contenttype is None:
            contenttype = guessMime(body) or 'application/octet-stream'
    
        maintype, subtype = contenttype.split('/', 1)
        if maintype == 'message':
            msg = MIMEBase(maintype, subtype)
            msg.set_payload(body)
            msg.add_header('Content-Disposition', 'inline')
        elif maintype == 'text':
            msg = MIMEText(body, _subtype=subtype)
        elif maintype == 'image':
            msg = MIMEImage(body, _subtype=subtype)
        elif maintype == 'audio':
            msg = MIMEAudio(body, _subtype=subtype)
        else:
            msg = MIMEBase(maintype, subtype)
            msg.set_payload(body)
            Encoders.encode_base64(msg)
        return msg