# -*- coding: utf-8 -*-
#
# Copyright (c) 2006-2007 by:
#     Blue Dynamics Alliance
#         * Klein & Partner KEG, Austria
#         * Squarewave Computing Robert Niederreiter, Austria
#
# GNU General Public License (GPL)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

__author__ = """Robert Niederreiter <rnix@squarewave.at>"""
__docformat__ = 'plaintext'

import os
import copy

from bda.email.imap.interfaces import IIMAPServerSettings
from bda.email.imap.interfaces import IIMAPProperties

from zope.interface import implements
from zope.component import adapts

from interfaces import IWebMail
from interfaces import IFsCacheProperties
from interfaces import ISMTPProperties
from interfaces import IWebmailProperties

from Products.CMFCore.utils import getToolByName


class IMAPServerSettings(object):
    """See bda.email.imap.interfaces.IServerSettings for detail.
    """
    
    implements(IIMAPServerSettings)
    adapts(IWebMail)
    
    def __init__(self, context):
        tool = context.portal_webmailtool
        self.folder_prefix = tool.getImapFolderPrefix()
        self.trash_folder = tool.getImapTrashFolder()
        self.sent_folder = tool.getImapSentFolder()
        self.draft_folder = tool.getImapDraftsFolder()
        self.show_prefix_option = tool.getImapShowPrefixOption()
        self.default_sub_of_inbox = tool.getImapDefaultSubOfInbox()
        self.show_contain_subfolders_option = \
            tool.getImapShowContainSubfoldersOption()
        self.delimiter = tool.getImapDelimiter()
        self.list_pattern = tool.getImapListPattern()
        self.delete_folder = tool.getImapDeleteFolder()
        self.force_username_lowercase = tool.getImapForceUsernameLowercase()
        self.folder_creation_encoding = tool.getFolderCreationEncoding()

class IMAPProperties(object):
    """See bda.email.imap.interfaces.IIMAPProperties for details
    """
    
    implements(IIMAPProperties)
    adapts(IWebMail)
    
    def __init__(self, context):
        self.context = context
    
    @property
    def server(self):
        """Return either optional IMAP server from mailbox object or default
        from webmailtool.
        """
        server = self.context.getOptionalImapServer().strip()
        if server is not '':
            return server
        return self.context.portal_webmailtool.getImapServer()
    
    @property
    def port(self):
        """Return either optional IMAP port from mailbox object or default
        from webmailtool.
        """
        port = self.context.getOptionalImapPort().strip()
        if port is not '':
            return port
        return self.context.portal_webmailtool.getImapPort()
    
    @property
    def user(self):
        """Return either username from PAS if corresponding flag set on mailbox
        or the username defined on the mailbox object.
        """
        if self.context.getUsePropertiesForAuthentication():
            pm = getToolByName(self.context, 'portal_membership')
            member = pm.getAuthenticatedMember()
            username = member.getUserName()
            return username
        else:
            return self.context.getUserName()
    
    @property
    def passwd(self):
        """Return either password from PAS (StoreAndRecallPassword) if
        corresponding flag set on mailbox or the password defined on the
        mailbox object.
        """
        if self.context.getUsePropertiesForAuthentication():
            pm = getToolByName(self.context, 'portal_membership')
            member = pm.getAuthenticatedMember()
            password = member.getProperty('password', '')
            return password
        else:
            return self.context.getPassword()


class SMTPProperties(object):
    """See interfaces.properties.ISMTPProperties for details
    """
    
    implements(ISMTPProperties)
    adapts(IWebMail)
    
    def __init__(self, context):
        self.context = context
        self.auth_required = context.getSmtpRequiresLogin()
    
    @property
    def server(self):
        """Return either optional SMTP server from mailbox object or default
        from webmailtool.
        """
        server = self.context.getOptionalSmtpServer().strip()
        if not server:
            return self.context.portal_webmailtool.getSmtpServer().strip()
        return server
    
    @property
    def port(self):
        """Return either optional SMTP port from mailbox object or default
        from webmailtool.
        """
        port = self.context.getOptionalSmtpPort().strip()
        if not port:
            return int(self.context.portal_webmailtool.getSmtpPort().strip())
        return int(port)
    
    @property
    def user(self):
        """Return either username from PAS if corresponding flag set on mailbox
        or the username defined on the mailbox object.
        """
        if self.context.getUsePropertiesForAuthentication():
            pm = getToolByName(self.context, 'portal_membership')
            member = pm.getAuthenticatedMember()
            username = member.getUserName()
            return username
        else:
            return self.context.getUserName().strip()
    
    @property
    def passwd(self):
        """Return either password from PAS (StoreAndRecallPassword) if
        corresponding flag set on mailbox or the password defined on the
        mailbox object.
        """
        if self.context.getUsePropertiesForAuthentication():
            pm = getToolByName(self.context, 'portal_membership')
            member = pm.getAuthenticatedMember()
            password = member.getProperty('password', '')
            return password
        else:
            return self.context.getPassword()


class FsCacheProperties(object):
    """See interfaces.properties.IFsCacheProperties for details
    """
    
    implements(IFsCacheProperties)
    adapts(IWebMail)
    
    def __init__(self, context):
        self.context = context
        self.timeout = context.getMailCacheTimeout()
        self.cache_enabled = self.timeout > 0 and True or False

    @property
    def cachedir(self):
        """Return the cachedir abs path.
        """
        if self.context.getUsePropertiesForAuthentication():
            pm = getToolByName(self.context, 'portal_membership')
            member = pm.getAuthenticatedMember()
            username = member.getUserName()
        else:
            username = self.context.getUserName()
        
        args = [
            globals()['__builtins__']['INSTANCE_HOME'],
            'var',
            'webmailcache',
            username.replace('.', '_').replace(' ', '_'),
        ]
        path = os.path.join(*args)
        return path
  

class WebmailProperties(object):
    """Interface to provide general webmail properties
    """
    
    implements(IWebmailProperties)
    adapts(IWebMail)
    
    def __init__(self, context):
        self.context = context
        self.max_messages = context.getMaxMessagesPerPage()
        self.real_name = context.getRealname()
        self.email_address = context.getFromAdress()
        self.create_format = context.getCreateFormat()
        self.actions = context.portal_actions.listFilteredActionsFor(context)
        self.message_part_information = context.getShowMessagePartInformation()
